<style type="text/css">{{d3_styles}}</style>

<body>
<script type="text/javascript">{{d3JS}}</script>
<script type="text/javascript">{{model_data}}</script>
<script type="text/javascript">{{draw}}</script>
</body>