body {
  font-family: Lato, sans-serif; }

h1.title, h1, h2, h3, h4, h5, h6, .h1, .h2, .h3, .h4, .h5, .h6 {
  font-family: Lato, sans-serif; }

button, .button {
  font-family: Lato, sans-serif; }

.media {
  padding: 24px;
  border: 1px solid rgba(0, 0, 0, 0.07);
  border-radius: 3px;
  margin-bottom: 24px;
  max-width: 400px;
  display: flex;
  align-items: flex-start; }
  .media img {
    margin: 4px 0; }
  .media .media-body {
    margin-left: 16px; }
    .media .media-body h5 {
      margin-bottom: 0; }
    .media .media-body p {
      margin-bottom: 0; }

.section-head {
  font-size: 24px;
  line-height: 32px;
  margin-top: 48px;
  font-weight: 900; }
  .section-head:after {
    content: '#';
    font-size: 15px;
    font-weight: normal;
    line-height: 1;
    color: rgba(0, 0, 0, 0.3);
    margin-left: 8px;
    position: relative;
    top: -1px; }
  .section-head a {
    text-decoration: none; }

.section-item-desc {
  font-family: Consolas, Monaco, "Courier New", monospace;
  font-size: 12px;
  color: rgba(0, 0, 0, 0.5); }

.example {
  border: 1px solid rgba(0, 0, 0, 0.07);
  padding: 32px;
  margin-bottom: 16px; }
  .example pre.code {
    margin-top: 40px;
    margin-bottom: 0;
    background: none;
    padding: 0; }
  .example.bg-darkgray {
    background: #313439; }
    .example.bg-darkgray pre.code {
      color: rgba(255, 255, 255, 0.85); }

.demo-muted-link,
.demo-muted-link:hover {
  text-decoration: none;
  color: rgba(0, 0, 0, 0.3); }

.demo-animation-wrap {
  margin-bottom: 24px; }
  .demo-animation-wrap:after {
    content: '';
    display: table;
    clear: both; }

.demo-animation-box {
  float: left;
  margin-right: 16px;
  width: 202px;
  height: 82px;
  border: 1px dashed rgba(0, 0, 0, 0.15); }
  .demo-animation-box > div {
    width: 200px;
    height: 80px;
    background: #f8f8f8;
    text-align: center;
    line-height: 80px;
    color: rgba(0, 0, 0, 0.4);
    font-size: 18px; }

.demo-animation-btn {
  font-size: 13px;
  text-transform: uppercase;
  font-weight: bold;
  display: inline-block;
  width: 200px;
  margin-right: 16px; }

.demo-sizing > div {
  font-family: Consolas, Monaco, "Courier New", monospace;
  font-size: 11px;
  padding-left: 4px;
  background: #d8e9fa;
  margin-bottom: 4px; }

.demo-grid .row {
  margin-bottom: 4px;
  background: #ebf4fc; }

.demo-grid .row.gutters {
  background: none; }

.demo-grid .col {
  font-family: Consolas, Monaco, "Courier New", monospace;
  font-size: 12px;
  padding: 8px 12px;
  background: #d8e9fa;
  border-left: 1px solid rgba(0, 0, 0, 0.1); }

.demo-grid .demo-col-nested {
  border-left: none;
  padding: 0; }
  .demo-grid .demo-col-nested .row {
    margin-bottom: 0; }

#demo-container {
  display: flex;
  flex-direction: row;
  flex-wrap: wrap; }

#demo-sidebar {
  flex: 0 0 300px;
  background: #c4def7; }

#demo-content {
  flex: auto;
  background: #ebf4fc; }

#demo-sidebar,
#demo-content {
  font-family: Consolas, Monaco, "Courier New", monospace;
  font-size: 12px;
  padding: 8px 12px;
  min-height: 80px; }

#demo-media-grid {
  -webkit-column-count: 2;
  -moz-column-count: 2;
  column-count: 2;
  -webkit-column-gap: 2%;
  -moz-column-gap: 2%;
  column-gap: 2%; }
  #demo-media-grid > div {
    display: inline-block;
    width: 100%; }
  @media (max-width: 768px) {
    #demo-media-grid {
      -webkit-column-count: 1;
      -moz-column-count: 1;
      column-count: 1; } }
  #demo-media-grid > div {
    font-family: Consolas, Monaco, "Courier New", monospace;
    font-size: 12px;
    padding: 8px 12px;
    background: #eae2f2;
    text-align: center;
    margin-bottom: 20px;
    height: 80px; }
  #demo-media-grid > div:nth-child(2n) {
    height: 200px; }
  #demo-media-grid > div:nth-child(5n) {
    height: 120px; }

.button.red {
  color: #fff;
  background-color: #ff3366; }
  .button.red:hover {
    color: #fff;
    background-color: #ff99b3; }
  .button.red:disabled, .button.red.disabled {
    color: rgba(255, 255, 255, 0.7);
    background-color: rgba(255, 51, 102, 0.7); }
  .button.red.outline {
    background: none;
    color: #ff3366;
    border-color: #ff3366; }
    .button.red.outline:hover {
      color: rgba(255, 51, 102, 0.6);
      border-color: rgba(255, 51, 102, 0.5); }
    .button.red.outline:disabled, .button.red.outline.disabled {
      background: none;
      color: rgba(255, 51, 102, 0.7);
      border-color: rgba(255, 51, 102, 0.5); }

.label.custom {
  background: #ea48a7;
  color: #fff; }
  .label.custom.tag, .label.custom.outline {
    background: none;
    border-color: #ea48a7;
    color: #ea48a7; }

#breadcrumbs-custom-separator li:after {
  content: '>'; }

.demo-gradient {
  height: 40px;
  margin-bottom: 24px; }

.demo-gradient-vertical {
  background-color: #5faac8;
  background-image: linear-gradient(to bottom, #5faac8 0%, #65ccb8 100%); }

.demo-gradient-vertical-to-opacity {
  background: linear-gradient(to bottom, #5faac8 0%, rgba(95, 170, 200, 0) 100%); }

.demo-gradient-horizontal {
  background-color: #5faac8;
  background: linear-gradient(to right, #5faac8 0%, #65ccb8 100%); }

.demo-gradient-horizontal-to-opacity {
  background: linear-gradient(to right, #5faac8 0%, rgba(95, 170, 200, 0) 100%); }

.demo-gradient-radial {
  background-image: radial-gradient(circle, #5faac8, #65ccb8); }

.example-inverted-box {
  display: inline-block;
  padding: 6px 8px 6px 8px;
  line-height: 1;
  vertical-align: middle;
  background: #d4d4d4; }

#livetabs {
  margin-bottom: 24px;
  font-size: 14px; }

#livetabs ul {
  display: flex; }

#livetabs a {
  color: #000;
  text-decoration: none;
  background: #f4f4f4;
  border-radius: 4px;
  padding: 4px 12px;
  border: 1px solid transparent; }
  #livetabs a:hover {
    opacity: .7; }

#livetabs li {
  margin-right: 4px; }

#livetabs li.active a {
  background: #fff;
  border-color: #eee;
  color: rgba(0, 0, 0, 0.5);
  cursor: default; }
  #livetabs li.active a:hover {
    opacity: 1; }

.togglebox-box {
  padding: 24px;
  padding-bottom: 16px;
  background: #f8f8f8;
  margin-bottom: 24px; }

#navbar-demo {
  display: flex;
  align-items: center;
  background: #f8f8f8;
  padding: 24px 20px;
  margin-bottom: 24px; }

#navbar-brand {
  margin-right: 24px; }

#navbar-main ul:after {
  content: '';
  display: table;
  clear: both; }

#navbar-main li {
  float: left;
  margin-right: 20px; }

#navbar-demo.fixed {
  background: rgba(255, 255, 255, 0.98);
  box-shadow: 0 1px 2px rgba(0, 0, 0, 0.2); }

#navbar-main li a {
  color: #000;
  text-decoration: none;
  display: block; }
  #navbar-main li a:hover {
    -moz-transition: all linear 0.2s;
    transition: all linear 0.2s;
    color: rgba(0, 0, 0, 0.6);
    text-decoration: underline; }

@media (max-width: 768px) {
  #navbar-demo {
    flex-direction: column;
    text-align: center; }
  #navbar-brand {
    margin: 0;
    margin-bottom: 20px; }
  #navbar-main li {
    float: none;
    margin: 0;
    margin-bottom: 20px; } }

#demo-nav-collapse,
#demo-nav-collapse ul {
  margin-left: 0;
  list-style: none; }

#demo-nav-collapse li {
  line-height: 32px; }

#demo-nav-collapse ul {
  margin-left: 20px;
  font-size: 14px; }

#demo-nav-collapse a {
  color: #000;
  text-decoration: none;
  display: block; }
  #demo-nav-collapse a:hover {
    -moz-transition: all linear 0.2s;
    transition: all linear 0.2s;
    color: rgba(0, 0, 0, 0.6);
    text-decoration: underline; }

.my-collapse {
  margin-bottom: 24px; }

.my-collapse h4 {
  background: #f4f4f4;
  padding: 8px 16px;
  margin-bottom: 1px;
  font-size: 15px;
  line-height: 24px; }

.my-collapse h4 a {
  text-decoration: none;
  color: #000;
  display: block; }

.my-collapse div {
  border: 1px solid rgba(0, 0, 0, 0.1);
  padding: 24px 32px 1px;
  margin-bottom: 1px; }

.swatch-box {
  text-align: center; }

.swatch-item {
  display: inline-block;
  margin: 24px; }
  .swatch-item h5 {
    font-family: Consolas, Monaco, "Courier New", monospace;
    font-weight: bold;
    font-size: 14px;
    line-height: 24px;
    margin-bottom: 0; }
  .swatch-item p {
    font-family: Consolas, Monaco, "Courier New", monospace;
    font-size: 12px;
    line-height: 20px;
    color: rgba(46, 47, 51, 0.65); }

.swatch {
  display: inline-block;
  height: 120px;
  width: 120px;
  border-radius: 120px;
  margin-bottom: 8px; }

.swatch-bg-headings {
  background: #0d0d0e; }

.swatch-bg-text {
  background: #313439; }

.swatch-bg-link {
  background: #007eff; }

.swatch-bg-link-hover {
  background: #ff3366; }

.swatch-bg-button-primary {
  background: #007eff; }

.swatch-bg-button-secondary {
  background: #313439; }

.swatch-bg-inverted {
  background: #fff; }

.swatch-bg-inverted {
  position: relative;
  bottom: -8px;
  margin-top: -8px;
  border: 8px solid #f8f8f8; }
